@@ -10,9 +10,10 @@ EXTRACTOR_HASDIGIT = "hasdigit"
 FEATURE_LAYER_WORD = "text"
 FEATURE_LAYER_LEMMA = "lemmas"
 FEATURE_LAYER_POS = "pos_tags"
+FEATURE_LAYER_GRAMMAR = "word_features"
 
 DEFAULT_EXTRACTORS = [EXTRACTOR_ISUPPER, EXTRACTOR_ISTITLE, EXTRACTOR_HASDIGIT]
-DEFAULT_LAYERS = [FEATURE_LAYER_WORD, FEATURE_LAYER_LEMMA, FEATURE_LAYER_POS]
+DEFAULT_LAYERS = [FEATURE_LAYER_WORD, FEATURE_LAYER_LEMMA, FEATURE_LAYER_POS, FEATURE_LAYER_GRAMMAR]
 
 
 def has_digit(string: str):
@@ -22,6 +23,18 @@ def has_digit(string: str):
     return any(chr.isdigit() for chr in string)
 
 
+def check_blacklist(string: str):
+    """
+    Checks whether token should be used as a feature.
+    Returns boolean.
+    """
+
+    # TODO: Implement
+
+    return False
+
+
+
 def extract_features(item: dict, layers_to_use: List[str], extractors_to_use: List[str], word_index: int, suffix_len: Tuple[int], embedding):
     """
     Extract features from token and surrounding tokens based on window size.
@@ -31,8 +44,18 @@ def extract_features(item: dict, layers_to_use: List[str], extractors_to_use: Li
     for layer in layers_to_use:
         if layer not in item:
             raise exceptions.InvalidInputError(f"Layer {layer} not present in document!")
-        value = item[layer].lower()
-        features.append(f"{word_index}:{layer}={value}")
+        # extract values from Grammar layer
+        if layer == FEATURE_LAYER_GRAMMAR:
+            for grammar_feature in item[layer].split("|"):
+                feat_list = grammar_feature.split("=")
+                # check if list contains key & val(otherwise it's useless)
+                if len(feat_list) == 2:
+                    features.append(f"{word_index}:{layer}:{feat_list[0]}={feat_list[1]}")
+        # extract values from other (flat) layers
+        else:
+            value = item[layer].lower()
+            if not check_blacklist(value):
+                features.append(f"{word_index}:{layer}={value}")
 
         # add some suffixes from token
         if layer == FEATURE_LAYER_WORD:
@@ -67,7 +90,11 @@ def extract_features(item: dict, layers_to_use: List[str], extractors_to_use: Li
     return features
 
 
-def word2features(sent: list, i: int, config, embedding):
+def word2features(sent: List[tuple], i: int, config, embedding):
+    """
+    Transforms token with it's layers into features.
+    I denotes the token ID in the list of tokens.
+    """
     # empty list for all features
     features = []
     # add bias
@@ -75,25 +102,16 @@ def word2features(sent: list, i: int, config, embedding):
         features.append("bias")
     # this is the current token
     item = {}
-    if FEATURE_LAYER_WORD in config.feature_layers:
-        item[FEATURE_LAYER_WORD] = sent[i][0]
-    if FEATURE_LAYER_LEMMA in config.feature_layers:
-        item[FEATURE_LAYER_LEMMA] = sent[i][1]
-    if FEATURE_LAYER_POS in config.feature_layers:
-        item[FEATURE_LAYER_POS] = sent[i][2]
-
+    for layer_name in config.feature_layers:
+        item[layer_name] = sent[i][0][layer_name]
     features.extend(extract_features(item, config.feature_layers, config.feature_extractors, "0", config.suffix_len, embedding))
     # check if not the first token in sentence
     if i > 0:
         for window in range(1, config.window_size+1):
             try:
                 item1 = {}
-                if FEATURE_LAYER_WORD in config.context_feature_layers:
-                    item1[FEATURE_LAYER_WORD] = sent[i-window][0]
-                if FEATURE_LAYER_LEMMA in config.context_feature_layers:
-                    item1[FEATURE_LAYER_LEMMA] = sent[i-window][1]
-                if FEATURE_LAYER_POS in config.context_feature_layers:
-                    item1[FEATURE_LAYER_POS] = sent[i-window][2]
+                for layer_name in config.context_feature_layers:
+                    item1[layer_name] = sent[i-window][0][layer_name]
                 features.extend(extract_features(item1, config.context_feature_layers, config.context_feature_extractors, f"-{window}", config.suffix_len, embedding))
             except IndexError:
                 pass
@@ -105,12 +123,8 @@ def word2features(sent: list, i: int, config, embedding):
         for window in range(1, config.window_size+1):
             try:
                 item1 = {}
-                if FEATURE_LAYER_WORD in config.context_feature_layers:
-                    item1[FEATURE_LAYER_WORD] = sent[i+window][0]
-                if FEATURE_LAYER_LEMMA in config.context_feature_layers:
-                    item1[FEATURE_LAYER_LEMMA] = sent[i+window][1]
-                if FEATURE_LAYER_POS in config.context_feature_layers:
-                    item1[FEATURE_LAYER_POS] = sent[i+window][2]
+                for layer_name in config.context_feature_layers:
+                    item1[layer_name] = sent[i+window][0][layer_name]
                 features.extend(extract_features(item1, config.context_feature_layers, config.context_feature_extractors, f"+{window}", config.suffix_len, embedding))
             except IndexError:
                 pass
@@ -120,8 +134,10 @@ def word2features(sent: list, i: int, config, embedding):
     return features
 
 
-def sent2features(sent: list, config, embedding):
+def sent2features(sent: List[tuple], config, embedding):
+    # featurize everything except labels
     return [word2features(sent, i, config, embedding) for i in range(len(sent))]
 
-def sent2labels(sent: list):
-    return [label for token, lemma, postag, label in sent]
+def sent2labels(sent: List[tuple]):
+    # retrieve labels from tuple
+    return [item[1] for item in sent]
